# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["SubscriptionCancelParams"]


class SubscriptionCancelParams(TypedDict, total=False):
    cancel_option: Required[Literal["end_of_subscription_term", "immediate", "requested_date"]]
    """Determines the timing of subscription cancellation"""

    cancellation_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """The date that the cancellation should take effect.

    This parameter can only be passed if the `cancel_option` is `requested_date`.
    """



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from .cost_list_params import CostListParams as CostListParams
from .cost_list_response import CostListResponse as CostListResponse
from .credit_list_params import CreditListParams as CreditListParams
from .usage_update_params import UsageUpdateParams as UsageUpdateParams
from .credit_list_response import CreditListResponse as CreditListResponse
from .usage_update_response import UsageUpdateResponse as UsageUpdateResponse
from .balance_transaction_list_params import (
    BalanceTransactionListParams as BalanceTransactionListParams,
)
from .cost_list_by_external_id_params import (
    CostListByExternalIDParams as CostListByExternalIDParams,
)
from .balance_transaction_create_params import (
    BalanceTransactionCreateParams as BalanceTransactionCreateParams,
)
from .balance_transaction_list_response import (
    BalanceTransactionListResponse as BalanceTransactionListResponse,
)
from .cost_list_by_external_id_response import (
    CostListByExternalIDResponse as CostListByExternalIDResponse,
)
from .credit_list_by_external_id_params import (
    CreditListByExternalIDParams as CreditListByExternalIDParams,
)
from .usage_update_by_external_id_params import (
    UsageUpdateByExternalIDParams as UsageUpdateByExternalIDParams,
)
from .balance_transaction_create_response import (
    BalanceTransactionCreateResponse as BalanceTransactionCreateResponse,
)
from .credit_list_by_external_id_response import (
    CreditListByExternalIDResponse as CreditListByExternalIDResponse,
)
from .usage_update_by_external_id_response import (
    UsageUpdateByExternalIDResponse as UsageUpdateByExternalIDResponse,
)



# File generated from our OpenAPI spec by Stainless.

from .prices import (
    Prices,
    AsyncPrices,
    PricesWithRawResponse,
    AsyncPricesWithRawResponse,
)
from .external_price_id import (
    ExternalPriceID,
    AsyncExternalPriceID,
    ExternalPriceIDWithRawResponse,
    AsyncExternalPriceIDWithRawResponse,
)

__all__ = [
    "ExternalPriceID",
    "AsyncExternalPriceID",
    "ExternalPriceIDWithRawResponse",
    "AsyncExternalPriceIDWithRawResponse",
    "Prices",
    "AsyncPrices",
    "PricesWithRawResponse",
    "AsyncPricesWithRawResponse",
]



# File generated from our OpenAPI spec by Stainless.

from .._models import BaseModel

__all__ = ["EventDeprecateResponse"]


class EventDeprecateResponse(BaseModel):
    deprecated: str
    """event_id of the deprecated event, if successfully updated"""



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Literal, Annotated, TypedDict

from ...._utils import PropertyInfo

__all__ = ["LedgerListParams"]


class LedgerListParams(TypedDict, total=False):
    created_at_gt: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[gt]", format="iso8601")]

    created_at_gte: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[gte]", format="iso8601")]

    created_at_lt: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[lt]", format="iso8601")]

    created_at_lte: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[lte]", format="iso8601")]

    currency: Optional[str]

    cursor: Optional[str]
    """Cursor for pagination.

    This can be populated by the `next_cursor` value returned from the initial
    request.
    """

    entry_status: Optional[Literal["committed", "pending"]]

    entry_type: Optional[
        Literal[
            "increment", "decrement", "expiration_change", "credit_block_expiry", "void", "void_initiated", "amendment"
        ]
    ]

    limit: int
    """The number of items to fetch. Defaults to 20."""

    minimum_amount: Optional[str]



# File generated from our OpenAPI spec by Stainless.

from typing import Dict, Union, Optional
from datetime import datetime
from typing_extensions import Literal

from ...._models import BaseModel

__all__ = [
    "LedgerListByExternalIDResponse",
    "IncrementLedgerEntry",
    "IncrementLedgerEntryCreditBlock",
    "IncrementLedgerEntryCustomer",
    "DecrementLedgerEntry",
    "DecrementLedgerEntryCreditBlock",
    "DecrementLedgerEntryCustomer",
    "ExpirationChangeLedgerEntry",
    "ExpirationChangeLedgerEntryCreditBlock",
    "ExpirationChangeLedgerEntryCustomer",
    "CreditBlockExpiryLedgerEntry",
    "CreditBlockExpiryLedgerEntryCreditBlock",
    "CreditBlockExpiryLedgerEntryCustomer",
    "VoidLedgerEntry",
    "VoidLedgerEntryCreditBlock",
    "VoidLedgerEntryCustomer",
    "VoidInitiatedLedgerEntry",
    "VoidInitiatedLedgerEntryCreditBlock",
    "VoidInitiatedLedgerEntryCustomer",
    "AmendmentLedgerEntry",
    "AmendmentLedgerEntryCreditBlock",
    "AmendmentLedgerEntryCustomer",
]


class IncrementLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class IncrementLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class IncrementLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: IncrementLedgerEntryCreditBlock

    customer: IncrementLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["increment"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    starting_balance: float


class DecrementLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class DecrementLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class DecrementLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: DecrementLedgerEntryCreditBlock

    customer: DecrementLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["decrement"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    starting_balance: float

    event_id: Optional[str] = None

    invoice_id: Optional[str] = None

    price_id: Optional[str] = None


class ExpirationChangeLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class ExpirationChangeLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class ExpirationChangeLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: ExpirationChangeLedgerEntryCreditBlock

    customer: ExpirationChangeLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["expiration_change"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    new_block_expiry_date: datetime

    starting_balance: float


class CreditBlockExpiryLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class CreditBlockExpiryLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class CreditBlockExpiryLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: CreditBlockExpiryLedgerEntryCreditBlock

    customer: CreditBlockExpiryLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["credit_block_expiry"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    starting_balance: float


class VoidLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class VoidLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class VoidLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: VoidLedgerEntryCreditBlock

    customer: VoidLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["void"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    starting_balance: float

    void_amount: float

    void_reason: Optional[str]


class VoidInitiatedLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class VoidInitiatedLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class VoidInitiatedLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: VoidInitiatedLedgerEntryCreditBlock

    customer: VoidInitiatedLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["void_initiated"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    new_block_expiry_date: datetime

    starting_balance: float

    void_amount: float

    void_reason: Optional[str]


class AmendmentLedgerEntryCreditBlock(BaseModel):
    id: str

    expiry_date: Optional[datetime]

    per_unit_cost_basis: Optional[str]


class AmendmentLedgerEntryCustomer(BaseModel):
    id: str

    external_customer_id: Optional[str]


class AmendmentLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: AmendmentLedgerEntryCreditBlock

    customer: AmendmentLedgerEntryCustomer

    description: Optional[str]

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["amendment"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    starting_balance: float


LedgerListByExternalIDResponse = Union[
    IncrementLedgerEntry,
    DecrementLedgerEntry,
    ExpirationChangeLedgerEntry,
    CreditBlockExpiryLedgerEntry,
    VoidLedgerEntry,
    VoidInitiatedLedgerEntry,
    AmendmentLedgerEntry,
]



# File generated from our OpenAPI spec by Stainless.

from .ledger import (
    Ledger,
    AsyncLedger,
    LedgerWithRawResponse,
    AsyncLedgerWithRawResponse,
)
from .credits import (
    Credits,
    AsyncCredits,
    CreditsWithRawResponse,
    AsyncCreditsWithRawResponse,
)

__all__ = [
    "Ledger",
    "AsyncLedger",
    "LedgerWithRawResponse",
    "AsyncLedgerWithRawResponse",
    "Credits",
    "AsyncCredits",
    "CreditsWithRawResponse",
    "AsyncCreditsWithRawResponse",
]



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import TYPE_CHECKING, Union, Optional
from datetime import datetime
from typing_extensions import Literal

import httpx

from ..._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ..._utils import maybe_transform
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_raw_response_wrapper, async_to_raw_response_wrapper
from ..._base_client import make_request_options
from ...types.customers import (
    CostListResponse,
    CostListByExternalIDResponse,
    cost_list_params,
    cost_list_by_external_id_params,
)

if TYPE_CHECKING:
    from ..._client import Orb, AsyncOrb

__all__ = ["Costs", "AsyncCosts"]


class Costs(SyncAPIResource):
    with_raw_response: CostsWithRawResponse

    def __init__(self, client: Orb) -> None:
        super().__init__(client)
        self.with_raw_response = CostsWithRawResponse(self)

    def list(
        self,
        customer_id: Optional[str],
        *,
        group_by: Optional[str] | NotGiven = NOT_GIVEN,
        timeframe_end: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        timeframe_start: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        view_mode: Optional[Literal["periodic", "cumulative"]] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> CostListResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](fetch-subscription-usage.api.mdx) to fetch usage
        per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](fetch-upcoming-invoice) resource, which represents a
           snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is $2.50, the customer's plan
        has a monthly minimum of $50 for this price, and that the subscription's billing
        period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | $22.50   | $50.00                   |
        | 2023-02-01      | 2023-02-03    | 19               | $47.50   | $50.00                   |
        | 2023-02-01      | 2023-02-04    | 20               | $50.00   | $50.00                   |
        | 2023-02-01      | 2023-02-05    | 28               | $70.00   | $70.00                   |
        | 2023-02-01      | 2023-02-06    | 36               | $90.00   | $90.00                   |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        If no timeframe bounds are specified, the response will default to the current
        billing period for the customer's subscription. For subscriptions that have
        ended, this will be the billing period when they were last active. If the
        subscription starts or ends within the timeframe, the response will only include
        windows where the subscription is active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ## Grouping by custom attributes

        In order to view costs grouped by a specific _attribute_ that each event is
        tagged with (i.e. `cluster`), you can additionally specify a `group_by` key. The
        `group_by` key denotes the event property on which to group.

        When returning grouped costs, a separate `price_group` object in the
        `per_price_costs` array is returned for each value of the `group_by` key present
        in your events. The `subtotal` value of the `per_price_costs` object is the sum
        of each `price_group`'s total.

        Orb expects events will contain values in the `properties` dictionary that
        correspond to the `group_by` key specified. By default, Orb will return a `null`
        group (i.e. events that match the metric but do not have the key set).
        Currently, it is only possible to view costs grouped by a single attribute at a
        time.

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          group_by: Groups per-price costs by the key provided.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get(
            f"/customers/{customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "group_by": group_by,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_params.CostListParams,
                ),
            ),
            cast_to=CostListResponse,
        )

    def list_by_external_id(
        self,
        external_customer_id: Optional[str],
        *,
        group_by: Optional[str] | NotGiven = NOT_GIVEN,
        timeframe_end: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        timeframe_start: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        view_mode: Optional[Literal["periodic", "cumulative"]] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> CostListByExternalIDResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](fetch-subscription-usage.api.mdx) to fetch usage
        per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](fetch-upcoming-invoice) resource, which represents a
           snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is $2.50, the customer's plan
        has a monthly minimum of $50 for this price, and that the subscription's billing
        period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | $22.50   | $50.00                   |
        | 2023-02-01      | 2023-02-03    | 19               | $47.50   | $50.00                   |
        | 2023-02-01      | 2023-02-04    | 20               | $50.00   | $50.00                   |
        | 2023-02-01      | 2023-02-05    | 28               | $70.00   | $70.00                   |
        | 2023-02-01      | 2023-02-06    | 36               | $90.00   | $90.00                   |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        If no timeframe bounds are specified, the response will default to the current
        billing period for the customer's subscription. For subscriptions that have
        ended, this will be the billing period when they were last active. If the
        subscription starts or ends within the timeframe, the response will only include
        windows where the subscription is active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ## Grouping by custom attributes

        In order to view costs grouped by a specific _attribute_ that each event is
        tagged with (i.e. `cluster`), you can additionally specify a `group_by` key. The
        `group_by` key denotes the event property on which to group.

        When returning grouped costs, a separate `price_group` object in the
        `per_price_costs` array is returned for each value of the `group_by` key present
        in your events. The `subtotal` value of the `per_price_costs` object is the sum
        of each `price_group`'s total.

        Orb expects events will contain values in the `properties` dictionary that
        correspond to the `group_by` key specified. By default, Orb will return a `null`
        group (i.e. events that match the metric but do not have the key set).
        Currently, it is only possible to view costs grouped by a single attribute at a
        time.

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          group_by: Groups per-price costs by the key provided.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get(
            f"/customers/external_customer_id/{external_customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "group_by": group_by,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_by_external_id_params.CostListByExternalIDParams,
                ),
            ),
            cast_to=CostListByExternalIDResponse,
        )


class AsyncCosts(AsyncAPIResource):
    with_raw_response: AsyncCostsWithRawResponse

    def __init__(self, client: AsyncOrb) -> None:
        super().__init__(client)
        self.with_raw_response = AsyncCostsWithRawResponse(self)

    async def list(
        self,
        customer_id: Optional[str],
        *,
        group_by: Optional[str] | NotGiven = NOT_GIVEN,
        timeframe_end: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        timeframe_start: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        view_mode: Optional[Literal["periodic", "cumulative"]] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> CostListResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](fetch-subscription-usage.api.mdx) to fetch usage
        per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](fetch-upcoming-invoice) resource, which represents a
           snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is $2.50, the customer's plan
        has a monthly minimum of $50 for this price, and that the subscription's billing
        period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | $22.50   | $50.00                   |
        | 2023-02-01      | 2023-02-03    | 19               | $47.50   | $50.00                   |
        | 2023-02-01      | 2023-02-04    | 20               | $50.00   | $50.00                   |
        | 2023-02-01      | 2023-02-05    | 28               | $70.00   | $70.00                   |
        | 2023-02-01      | 2023-02-06    | 36               | $90.00   | $90.00                   |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        If no timeframe bounds are specified, the response will default to the current
        billing period for the customer's subscription. For subscriptions that have
        ended, this will be the billing period when they were last active. If the
        subscription starts or ends within the timeframe, the response will only include
        windows where the subscription is active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ## Grouping by custom attributes

        In order to view costs grouped by a specific _attribute_ that each event is
        tagged with (i.e. `cluster`), you can additionally specify a `group_by` key. The
        `group_by` key denotes the event property on which to group.

        When returning grouped costs, a separate `price_group` object in the
        `per_price_costs` array is returned for each value of the `group_by` key present
        in your events. The `subtotal` value of the `per_price_costs` object is the sum
        of each `price_group`'s total.

        Orb expects events will contain values in the `properties` dictionary that
        correspond to the `group_by` key specified. By default, Orb will return a `null`
        group (i.e. events that match the metric but do not have the key set).
        Currently, it is only possible to view costs grouped by a single attribute at a
        time.

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          group_by: Groups per-price costs by the key provided.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._get(
            f"/customers/{customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "group_by": group_by,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_params.CostListParams,
                ),
            ),
            cast_to=CostListResponse,
        )

    async def list_by_external_id(
        self,
        external_customer_id: Optional[str],
        *,
        group_by: Optional[str] | NotGiven = NOT_GIVEN,
        timeframe_end: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        timeframe_start: Union[str, datetime, None] | NotGiven = NOT_GIVEN,
        view_mode: Optional[Literal["periodic", "cumulative"]] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> CostListByExternalIDResponse:
        """
        This endpoint is used to fetch a day-by-day snapshot of a customer's costs in
        Orb, calculated by applying pricing information to the underlying usage (see the
        [subscription usage endpoint](fetch-subscription-usage.api.mdx) to fetch usage
        per metric, in usage units rather than a currency).

        This endpoint can be leveraged for internal tooling and to provide a more
        transparent billing experience for your end users:

        1. Understand the cost breakdown per line item historically and in real-time for
           the current billing period.
        2. Provide customer visibility into how different services are contributing to
           the overall invoice with a per-day timeseries (as compared to the
           [upcoming invoice](fetch-upcoming-invoice) resource, which represents a
           snapshot for the current period).
        3. Assess how minimums and discounts affect your customers by teasing apart
           costs directly as a result of usage, as opposed to minimums and discounts at
           the plan and price level.
        4. Gain insight into key customer health metrics, such as the percent
           utilization of the minimum committed spend.

        ## Fetching subscriptions

        By default, this endpoint fetches the currently active subscription for the
        customer, and returns cost information for the subscription's current billing
        period, broken down by each participating price. If there are no currently
        active subscriptions, this will instead default to the most recently active
        subscription or return an empty series if none are found. For example, if your
        plan charges for compute hours, job runs, and data syncs, then this endpoint
        would provide a daily breakdown of your customer's cost for each of those axes.

        If timeframe bounds are specified, Orb fetches all subscriptions that were
        active in that timeframe. If two subscriptions overlap on a single day, costs
        from each price will be summed, and prices for both subscriptions will be
        included in the breakdown.

        ## Prepaid plans

        For plans that include prices which deduct credits rather than accrue in-arrears
        charges in a billable currency, this endpoint will return the total deduction
        amount, in credits, for the specified timeframe.

        ## Cumulative subtotals and totals

        Since the subtotal and total must factor in any billing-period level discounts
        and minimums, it's most meaningful to consider costs relative to the start of
        the subscription's billing period. As a result, by default this endpoint returns
        cumulative totals since the beginning of the billing period. In particular, the
        `timeframe_start` of a returned timeframe window is _always_ the beginning of
        the billing period and `timeframe_end` is incremented one day at a time to build
        the result.

        A customer that uses a few API calls a day but has a minimum commitment might
        exhibit the following pattern for their subtotal and total in the first few days
        of the month. Here, we assume that each API call is $2.50, the customer's plan
        has a monthly minimum of $50 for this price, and that the subscription's billing
        period bounds are aligned to the first of the month:

        | timeframe_start | timeframe_end | Cumulative usage | Subtotal | Total (incl. commitment) |
        | --------------- | ------------- | ---------------- | -------- | ------------------------ |
        | 2023-02-01      | 2023-02-02    | 9                | $22.50   | $50.00                   |
        | 2023-02-01      | 2023-02-03    | 19               | $47.50   | $50.00                   |
        | 2023-02-01      | 2023-02-04    | 20               | $50.00   | $50.00                   |
        | 2023-02-01      | 2023-02-05    | 28               | $70.00   | $70.00                   |
        | 2023-02-01      | 2023-02-06    | 36               | $90.00   | $90.00                   |

        ### Periodic values

        When the query parameter `view_mode=periodic` is specified, Orb will return an
        incremental day-by-day view of costs. In this case, there will always be a
        one-day difference between `timeframe_start` and `timeframe_end` for the
        timeframes returned. This is a transform on top of the cumulative costs,
        calculated by taking the difference of each timeframe with the last. Note that
        in the above example, the `Total` value would be 0 for the second two data
        points, since the minimum commitment has not yet been hit and each day is not
        contributing anything to the total cost.

        ## Timeframe bounds

        If no timeframe bounds are specified, the response will default to the current
        billing period for the customer's subscription. For subscriptions that have
        ended, this will be the billing period when they were last active. If the
        subscription starts or ends within the timeframe, the response will only include
        windows where the subscription is active.

        As noted above, `timeframe_start` for a given cumulative datapoint is always the
        beginning of the billing period, and `timeframe_end` is incremented one day at a
        time to construct the response. When a timeframe is passed in that is not
        aligned to the current subscription's billing period, the response will contain
        cumulative totals from multiple billing periods.

        Suppose the queried customer has a subscription aligned to the 15th of every
        month. If this endpoint is queried with the date range `2023-06-01` -
        `2023-07-01`, the first data point will represent about half a billing period's
        worth of costs, accounting for accruals from the start of the billing period and
        inclusive of the first day of the timeframe
        (`timeframe_start = 2023-05-15 00:00:00`, `timeframe_end = 2023-06-02 00:00:00`)

        | datapoint index | timeframe_start | timeframe_end |
        | --------------- | --------------- | ------------- |
        | 0               | 2023-05-15      | 2023-06-02    |
        | 1               | 2023-05-15      | 2023-06-03    |
        | 2               | ...             | ...           |
        | 3               | 2023-05-15      | 2023-06-14    |
        | 4               | 2023-06-15      | 2023-06-16    |
        | 5               | 2023-06-15      | 2023-06-17    |
        | 6               | ...             | ...           |
        | 7               | 2023-06-15      | 2023-07-01    |

        You can see this sliced timeframe visualized
        [here](https://i.imgur.com/TXhYgme.png).

        ## Grouping by custom attributes

        In order to view costs grouped by a specific _attribute_ that each event is
        tagged with (i.e. `cluster`), you can additionally specify a `group_by` key. The
        `group_by` key denotes the event property on which to group.

        When returning grouped costs, a separate `price_group` object in the
        `per_price_costs` array is returned for each value of the `group_by` key present
        in your events. The `subtotal` value of the `per_price_costs` object is the sum
        of each `price_group`'s total.

        Orb expects events will contain values in the `properties` dictionary that
        correspond to the `group_by` key specified. By default, Orb will return a `null`
        group (i.e. events that match the metric but do not have the key set).
        Currently, it is only possible to view costs grouped by a single attribute at a
        time.

        ### Matrix prices

        When a price uses matrix pricing, it's important to view costs grouped by those
        matrix dimensions. Orb will return `price_groups` with the `grouping_key` and
        `secondary_grouping_key` based on the matrix price definition, for each
        `grouping_value` and `secondary_grouping_value` available.

        Args:
          group_by: Groups per-price costs by the key provided.

          timeframe_end: Costs returned are exclusive of `timeframe_end`.

          timeframe_start: Costs returned are inclusive of `timeframe_start`.

          view_mode: Controls whether Orb returns cumulative costs since the start of the billing
              period, or incremental day-by-day costs. If your customer has minimums or
              discounts, it's strongly recommended that you use the default cumulative
              behavior.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._get(
            f"/customers/external_customer_id/{external_customer_id}/costs",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "group_by": group_by,
                        "timeframe_end": timeframe_end,
                        "timeframe_start": timeframe_start,
                        "view_mode": view_mode,
                    },
                    cost_list_by_external_id_params.CostListByExternalIDParams,
                ),
            ),
            cast_to=CostListByExternalIDResponse,
        )


class CostsWithRawResponse:
    def __init__(self, costs: Costs) -> None:
        self.list = to_raw_response_wrapper(
            costs.list,
        )
        self.list_by_external_id = to_raw_response_wrapper(
            costs.list_by_external_id,
        )


class AsyncCostsWithRawResponse:
    def __init__(self, costs: AsyncCosts) -> None:
        self.list = async_to_raw_response_wrapper(
            costs.list,
        )
        self.list_by_external_id = async_to_raw_response_wrapper(
            costs.list_by_external_id,
        )



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import httpx

from ...types import Subscription
from ..._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ..._utils import maybe_transform
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_raw_response_wrapper, async_to_raw_response_wrapper
from ...pagination import SyncPage, AsyncPage
from ..._base_client import AsyncPaginator, make_request_options
from ...types.coupons import subscription_list_params

if TYPE_CHECKING:
    from ..._client import Orb, AsyncOrb

__all__ = ["Subscriptions", "AsyncSubscriptions"]


class Subscriptions(SyncAPIResource):
    with_raw_response: SubscriptionsWithRawResponse

    def __init__(self, client: Orb) -> None:
        super().__init__(client)
        self.with_raw_response = SubscriptionsWithRawResponse(self)

    def list(
        self,
        coupon_id: str,
        *,
        cursor: Optional[str] | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncPage[Subscription]:
        """
        This endpoint returns a list of all subscriptions that have redeemed a given
        coupon as a [paginated](../reference/pagination) list, ordered starting from the
        most recently created subscription. For a full discussion of the subscription
        resource, see [Subscription](../guides/concepts#subscription).

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            f"/coupons/{coupon_id}/subscriptions",
            page=SyncPage[Subscription],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "limit": limit,
                    },
                    subscription_list_params.SubscriptionListParams,
                ),
            ),
            model=Subscription,
        )


class AsyncSubscriptions(AsyncAPIResource):
    with_raw_response: AsyncSubscriptionsWithRawResponse

    def __init__(self, client: AsyncOrb) -> None:
        super().__init__(client)
        self.with_raw_response = AsyncSubscriptionsWithRawResponse(self)

    def list(
        self,
        coupon_id: str,
        *,
        cursor: Optional[str] | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[Subscription, AsyncPage[Subscription]]:
        """
        This endpoint returns a list of all subscriptions that have redeemed a given
        coupon as a [paginated](../reference/pagination) list, ordered starting from the
        most recently created subscription. For a full discussion of the subscription
        resource, see [Subscription](../guides/concepts#subscription).

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            f"/coupons/{coupon_id}/subscriptions",
            page=AsyncPage[Subscription],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "limit": limit,
                    },
                    subscription_list_params.SubscriptionListParams,
                ),
            ),
            model=Subscription,
        )


class SubscriptionsWithRawResponse:
    def __init__(self, subscriptions: Subscriptions) -> None:
        self.list = to_raw_response_wrapper(
            subscriptions.list,
        )


class AsyncSubscriptionsWithRawResponse:
    def __init__(self, subscriptions: AsyncSubscriptions) -> None:
        self.list = async_to_raw_response_wrapper(
            subscriptions.list,
        )



# File generated from our OpenAPI spec by Stainless.

from typing import Dict, List, Optional
from datetime import datetime
from typing_extensions import Literal

from .price import Price
from .shared import Discount
from .._models import BaseModel

__all__ = [
    "Plan",
    "BasePlan",
    "Maximum",
    "Minimum",
    "PlanPhase",
    "PlanPhaseMaximum",
    "PlanPhaseMinimum",
    "Product",
    "TrialConfig",
]


class BasePlan(BaseModel):
    id: Optional[str]

    external_plan_id: Optional[str]
    """
    An optional user-defined ID for this plan resource, used throughout the system
    as an alias for this Plan. Use this field to identify a plan by an existing
    identifier in your system.
    """

    name: Optional[str]


class Maximum(BaseModel):
    applies_to_price_ids: List[str]
    """List of price_ids that this maximum amount applies to.

    For plan/plan phase maximums, this can be a subset of prices.
    """

    maximum_amount: str
    """Maximum amount applied"""


class Minimum(BaseModel):
    applies_to_price_ids: List[str]
    """List of price_ids that this minimum amount applies to.

    For plan/plan phase minimums, this can be a subset of prices.
    """

    minimum_amount: str
    """Minimum amount applied"""


class PlanPhaseMaximum(BaseModel):
    applies_to_price_ids: List[str]
    """List of price_ids that this maximum amount applies to.

    For plan/plan phase maximums, this can be a subset of prices.
    """

    maximum_amount: str
    """Maximum amount applied"""


class PlanPhaseMinimum(BaseModel):
    applies_to_price_ids: List[str]
    """List of price_ids that this minimum amount applies to.

    For plan/plan phase minimums, this can be a subset of prices.
    """

    minimum_amount: str
    """Minimum amount applied"""


class PlanPhase(BaseModel):
    id: str

    description: Optional[str]

    discount: Optional[Discount]

    duration: Optional[int]
    """How many terms of length `duration_unit` this phase is active for.

    If null, this phase is evergreen and active indefinitely
    """

    duration_unit: Optional[Literal["daily", "monthly", "quarterly", "annual"]]

    maximum: Optional[PlanPhaseMaximum]

    maximum_amount: Optional[str]

    minimum: Optional[PlanPhaseMinimum]

    minimum_amount: Optional[str]

    name: str

    order: int
    """Determines the ordering of the phase in a plan's lifecycle. 1 = first phase."""


class Product(BaseModel):
    id: str

    created_at: datetime

    name: str


class TrialConfig(BaseModel):
    trial_period: Optional[int]

    trial_period_unit: Literal["days"]


class Plan(BaseModel):
    id: str

    base_plan: Optional[BasePlan]

    base_plan_id: Optional[str]
    """
    The parent plan id if the given plan was created by overriding one or more of
    the parent's prices
    """

    created_at: datetime

    currency: str
    """
    An ISO 4217 currency string or custom pricing unit (`credits`) for this plan's
    prices.
    """

    default_invoice_memo: Optional[str]
    """
    The default memo text on the invoices corresponding to subscriptions on this
    plan. Note that each subscription may configure its own memo.
    """

    description: str

    discount: Optional[Discount]

    external_plan_id: Optional[str]
    """
    An optional user-defined ID for this plan resource, used throughout the system
    as an alias for this Plan. Use this field to identify a plan by an existing
    identifier in your system.
    """

    invoicing_currency: str
    """An ISO 4217 currency string for which this plan is billed in.

    Matches `currency` unless `currency` is a custom pricing unit.
    """

    maximum: Optional[Maximum]

    maximum_amount: Optional[str]

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    minimum: Optional[Minimum]

    minimum_amount: Optional[str]

    name: str

    net_terms: Optional[int]
    """Determines the difference between the invoice issue date and the due date.

    A value of "0" here signifies that invoices are due on issue, whereas a value of
    "30" means that the customer has a month to pay the invoice before its overdue.
    Note that individual subscriptions or invoices may set a different net terms
    configuration.
    """

    plan_phases: Optional[List[PlanPhase]]

    prices: List[Price]
    """Prices for this plan.

    If the plan has phases, this includes prices across all phases of the plan.
    """

    product: Product

    status: Literal["active", "archived", "draft"]

    trial_config: TrialConfig



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import List, Union, Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["CouponCreateParams", "Discount", "DiscountPercentageDiscount", "DiscountAmountDiscount"]


class CouponCreateParams(TypedDict, total=False):
    discount: Required[Discount]

    redemption_code: Required[str]
    """This string can be used to redeem this coupon for a given subscription."""

    duration_in_months: Optional[int]
    """
    This allows for a coupon's discount to apply for a limited time (determined in
    months); a `null` value here means "unlimited time".
    """

    max_redemptions: Optional[int]
    """
    The maximum number of redemptions allowed for this coupon before it is
    exhausted;`null` here means "unlimited".
    """


class DiscountPercentageDiscount(TypedDict, total=False):
    applies_to_price_ids: Required[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    discount_type: Required[Literal["percentage"]]

    percentage_discount: Required[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    reason: Optional[str]


class DiscountAmountDiscount(TypedDict, total=False):
    amount_discount: Required[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Required[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    discount_type: Required[Literal["amount"]]

    reason: Optional[str]


Discount = Union[DiscountPercentageDiscount, DiscountAmountDiscount]



# File generated from our OpenAPI spec by Stainless.

from typing import List, Optional

from .._models import BaseModel
from .subscription import Subscription

__all__ = ["Subscriptions", "PaginationMetadata"]


class PaginationMetadata(BaseModel):
    has_more: bool

    next_cursor: Optional[str]


class Subscriptions(BaseModel):
    data: List[Subscription]

    pagination_metadata: PaginationMetadata



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import httpx

from ..types import Item, item_list_params, item_create_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import maybe_transform
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import to_raw_response_wrapper, async_to_raw_response_wrapper
from ..pagination import SyncPage, AsyncPage
from .._base_client import AsyncPaginator, make_request_options

if TYPE_CHECKING:
    from .._client import Orb, AsyncOrb

__all__ = ["Items", "AsyncItems"]


class Items(SyncAPIResource):
    with_raw_response: ItemsWithRawResponse

    def __init__(self, client: Orb) -> None:
        super().__init__(client)
        self.with_raw_response = ItemsWithRawResponse(self)

    def create(
        self,
        *,
        name: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> Item:
        """
        This endpoint is used to create an [Item](../guides/concepts#item).

        Args:
          name: The name of the item.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/items",
            body=maybe_transform({"name": name}, item_create_params.ItemCreateParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Item,
        )

    def list(
        self,
        *,
        cursor: Optional[str] | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncPage[Item]:
        """
        This endpoint returns a list of all Items, ordered in descending order by
        creation time.

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/items",
            page=SyncPage[Item],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "limit": limit,
                    },
                    item_list_params.ItemListParams,
                ),
            ),
            model=Item,
        )

    def fetch(
        self,
        item_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> Item:
        """
        This endpoint returns an item identified by its item_id.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get(
            f"/items/{item_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Item,
        )


class AsyncItems(AsyncAPIResource):
    with_raw_response: AsyncItemsWithRawResponse

    def __init__(self, client: AsyncOrb) -> None:
        super().__init__(client)
        self.with_raw_response = AsyncItemsWithRawResponse(self)

    async def create(
        self,
        *,
        name: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> Item:
        """
        This endpoint is used to create an [Item](../guides/concepts#item).

        Args:
          name: The name of the item.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/items",
            body=maybe_transform({"name": name}, item_create_params.ItemCreateParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Item,
        )

    def list(
        self,
        *,
        cursor: Optional[str] | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[Item, AsyncPage[Item]]:
        """
        This endpoint returns a list of all Items, ordered in descending order by
        creation time.

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/items",
            page=AsyncPage[Item],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "limit": limit,
                    },
                    item_list_params.ItemListParams,
                ),
            ),
            model=Item,
        )

    async def fetch(
        self,
        item_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> Item:
        """
        This endpoint returns an item identified by its item_id.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._get(
            f"/items/{item_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Item,
        )


class ItemsWithRawResponse:
    def __init__(self, items: Items) -> None:
        self.create = to_raw_response_wrapper(
            items.create,
        )
        self.list = to_raw_response_wrapper(
            items.list,
        )
        self.fetch = to_raw_response_wrapper(
            items.fetch,
        )


class AsyncItemsWithRawResponse:
    def __init__(self, items: AsyncItems) -> None:
        self.create = async_to_raw_response_wrapper(
            items.create,
        )
        self.list = async_to_raw_response_wrapper(
            items.list,
        )
        self.fetch = async_to_raw_response_wrapper(
            items.fetch,
        )



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["MetricListParams"]


class MetricListParams(TypedDict, total=False):
    created_at_gt: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[gt]", format="iso8601")]

    created_at_gte: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[gte]", format="iso8601")]

    created_at_lt: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[lt]", format="iso8601")]

    created_at_lte: Annotated[Union[str, datetime, None], PropertyInfo(alias="created_at[lte]", format="iso8601")]

    cursor: Optional[str]
    """Cursor for pagination.

    This can be populated by the `next_cursor` value returned from the initial
    request.
    """

    limit: int
    """The number of items to fetch. Defaults to 20."""



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import List, Union, Optional
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["EventIngestParams", "Event"]


class EventIngestParams(TypedDict, total=False):
    events: Required[List[Event]]

    backfill_id: Optional[str]
    """
    If this ingestion request is part of a backfill, this parameter ties the
    ingested events to the backfill
    """

    debug: bool
    """Flag to enable additional debug information in the endpoint response"""


class Event(TypedDict, total=False):
    event_name: Required[str]
    """A name to meaningfully identify the action or event type."""

    idempotency_key: Required[str]
    """A unique value, generated by the client, that is used to de-duplicate events.

    Exactly one event with a given idempotency key will be ingested, which allows
    for safe request retries.
    """

    properties: Required[object]
    """A dictionary of custom properties.

    Values in this dictionary must be numeric, boolean, or strings. Nested
    dictionaries are disallowed.
    """

    timestamp: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """An ISO 8601 format date with no timezone offset (i.e.

    UTC). This should represent the time that usage was recorded, and is
    particularly important to attribute usage to a given billing period.
    """

    customer_id: Optional[str]
    """The Orb Customer identifier"""

    external_customer_id: Optional[str]
    """
    An alias for the Orb customer, whose mapping is specified when creating the
    customer
    """



from __future__ import annotations

from os import PathLike
from abc import ABC, abstractmethod
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Type,
    Tuple,
    Union,
    Mapping,
    TypeVar,
    Callable,
    Iterator,
    Optional,
    Sequence,
    AsyncIterator,
)
from typing_extensions import (
    Literal,
    Protocol,
    TypeAlias,
    TypedDict,
    override,
    runtime_checkable,
)

import pydantic
from httpx import URL, Proxy, Timeout, Response, BaseTransport, AsyncBaseTransport

if TYPE_CHECKING:
    from ._models import BaseModel

Transport = BaseTransport
AsyncTransport = AsyncBaseTransport
Query = Mapping[str, object]
Body = object
AnyMapping = Mapping[str, object]
ModelT = TypeVar("ModelT", bound=pydantic.BaseModel)
_T = TypeVar("_T")


class BinaryResponseContent(ABC):
    def __init__(
        self,
        response: Any,
    ) -> None:
        ...

    @property
    @abstractmethod
    def content(self) -> bytes:
        pass

    @property
    @abstractmethod
    def text(self) -> str:
        pass

    @property
    @abstractmethod
    def encoding(self) -> Optional[str]:
        """
        Return an encoding to use for decoding the byte content into text.
        The priority for determining this is given by...

        * `.encoding = <>` has been set explicitly.
        * The encoding as specified by the charset parameter in the Content-Type header.
        * The encoding as determined by `default_encoding`, which may either be
          a string like "utf-8" indicating the encoding to use, or may be a callable
          which enables charset autodetection.
        """
        pass

    @property
    @abstractmethod
    def charset_encoding(self) -> Optional[str]:
        """
        Return the encoding, as specified by the Content-Type header.
        """
        pass

    @abstractmethod
    def json(self, **kwargs: Any) -> Any:
        pass

    @abstractmethod
    def read(self) -> bytes:
        """
        Read and return the response content.
        """
        pass

    @abstractmethod
    def iter_bytes(self, chunk_size: Optional[int] = None) -> Iterator[bytes]:
        """
        A byte-iterator over the decoded response content.
        This allows us to handle gzip, deflate, and brotli encoded responses.
        """
        pass

    @abstractmethod
    def iter_text(self, chunk_size: Optional[int] = None) -> Iterator[str]:
        """
        A str-iterator over the decoded response content
        that handles both gzip, deflate, etc but also detects the content's
        string encoding.
        """
        pass

    @abstractmethod
    def iter_lines(self) -> Iterator[str]:
        pass

    @abstractmethod
    def iter_raw(self, chunk_size: Optional[int] = None) -> Iterator[bytes]:
        """
        A byte-iterator over the raw response content.
        """
        pass

    @abstractmethod
    def stream_to_file(
        self,
        file: str | PathLike[str],
        *,
        chunk_size: int | None = None,
    ) -> None:
        """
        Stream the output to the given file.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the response and release the connection.
        Automatically called if the response body is read to completion.
        """
        pass

    @abstractmethod
    async def aread(self) -> bytes:
        """
        Read and return the response content.
        """
        pass

    @abstractmethod
    async def aiter_bytes(self, chunk_size: Optional[int] = None) -> AsyncIterator[bytes]:
        """
        A byte-iterator over the decoded response content.
        This allows us to handle gzip, deflate, and brotli encoded responses.
        """
        pass

    @abstractmethod
    async def aiter_text(self, chunk_size: Optional[int] = None) -> AsyncIterator[str]:
        """
        A str-iterator over the decoded response content
        that handles both gzip, deflate, etc but also detects the content's
        string encoding.
        """
        pass

    @abstractmethod
    async def aiter_lines(self) -> AsyncIterator[str]:
        pass

    @abstractmethod
    async def aiter_raw(self, chunk_size: Optional[int] = None) -> AsyncIterator[bytes]:
        """
        A byte-iterator over the raw response content.
        """
        pass

    @abstractmethod
    async def astream_to_file(
        self,
        file: str | PathLike[str],
        *,
        chunk_size: int | None = None,
    ) -> None:
        """
        Stream the output to the given file.
        """
        pass

    @abstractmethod
    async def aclose(self) -> None:
        """
        Close the response and release the connection.
        Automatically called if the response body is read to completion.
        """
        pass


# Approximates httpx internal ProxiesTypes and RequestFiles types
# while adding support for `PathLike` instances
ProxiesDict = Dict["str | URL", Union[None, str, URL, Proxy]]
ProxiesTypes = Union[str, Proxy, ProxiesDict]
if TYPE_CHECKING:
    FileContent = Union[IO[bytes], bytes, PathLike[str]]
else:
    FileContent = Union[IO[bytes], bytes, PathLike]  # PathLike is not subscriptable in Python 3.8.
FileTypes = Union[
    # file (or bytes)
    FileContent,
    # (filename, file (or bytes))
    Tuple[Optional[str], FileContent],
    # (filename, file (or bytes), content_type)
    Tuple[Optional[str], FileContent, Optional[str]],
    # (filename, file (or bytes), content_type, headers)
    Tuple[Optional[str], FileContent, Optional[str], Mapping[str, str]],
]
RequestFiles = Union[Mapping[str, FileTypes], Sequence[Tuple[str, FileTypes]]]

# duplicate of the above but without our custom file support
HttpxFileContent = Union[IO[bytes], bytes]
HttpxFileTypes = Union[
    # file (or bytes)
    HttpxFileContent,
    # (filename, file (or bytes))
    Tuple[Optional[str], HttpxFileContent],
    # (filename, file (or bytes), content_type)
    Tuple[Optional[str], HttpxFileContent, Optional[str]],
    # (filename, file (or bytes), content_type, headers)
    Tuple[Optional[str], HttpxFileContent, Optional[str], Mapping[str, str]],
]
HttpxRequestFiles = Union[Mapping[str, HttpxFileTypes], Sequence[Tuple[str, HttpxFileTypes]]]

# Workaround to support (cast_to: Type[ResponseT]) -> ResponseT
# where ResponseT includes `None`. In order to support directly
# passing `None`, overloads would have to be defined for every
# method that uses `ResponseT` which would lead to an unacceptable
# amount of code duplication and make it unreadable. See _base_client.py
# for example usage.
#
# This unfortunately means that you will either have
# to import this type and pass it explicitly:
#
# from orb import NoneType
# client.get('/foo', cast_to=NoneType)
#
# or build it yourself:
#
# client.get('/foo', cast_to=type(None))
if TYPE_CHECKING:
    NoneType: Type[None]
else:
    NoneType = type(None)


class RequestOptions(TypedDict, total=False):
    headers: Headers
    max_retries: int
    timeout: float | Timeout | None
    params: Query
    extra_json: AnyMapping
    idempotency_key: str


# Sentinel class used when the response type is an object with an unknown schema
class UnknownResponse:
    ...


# Sentinel class used until PEP 0661 is accepted
class NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None (which may have different behavior).

    For example:

    ```py
    def get(timeout: Union[int, NotGiven, None] = NotGiven()) -> Response: ...

    get(timeout=1) # 1s timeout
    get(timeout=None) # No timeout
    get() # Default timeout behavior, which may not be statically known at the method definition.
    ```
    """

    def __bool__(self) -> Literal[False]:
        return False

    @override
    def __repr__(self) -> str:
        return "NOT_GIVEN"


NotGivenOr = Union[_T, NotGiven]
NOT_GIVEN = NotGiven()


class Omit:
    """In certain situations you need to be able to represent a case where a default value has
    to be explicitly removed and `None` is not an appropriate substitute, for example:

    ```py
    # as the default `Content-Type` header is `application/json` that will be sent
    client.post('/upload/files', files={'file': b'my raw file content'})

    # you can't explicitly override the header as it has to be dynamically generated
    # to look something like: 'multipart/form-data; boundary=0d8382fcf5f8c3be01ca2e11002d2983'
    client.post(..., headers={'Content-Type': 'multipart/form-data'})

    # instead you can remove the default `application/json` header by passing Omit
    client.post(..., headers={'Content-Type': Omit()})
    ```
    """

    def __bool__(self) -> Literal[False]:
        return False


@runtime_checkable
class ModelBuilderProtocol(Protocol):
    @classmethod
    def build(
        cls: type[_T],
        *,
        response: Response,
        data: object,
    ) -> _T:
        ...


Headers = Mapping[str, Union[str, Omit]]


class HeadersLikeProtocol(Protocol):
    def get(self, __key: str) -> str | None:
        ...


HeadersLike = Union[Headers, HeadersLikeProtocol]

ResponseT = TypeVar(
    "ResponseT",
    bound="Union[str, None, BaseModel, List[Any], Dict[str, Any], Response, UnknownResponse, ModelBuilderProtocol, BinaryResponseContent]",
)

StrBytesIntFloat = Union[str, bytes, int, float]

# Note: copied from Pydantic
# https://github.com/pydantic/pydantic/blob/32ea570bf96e84234d2992e1ddf40ab8a565925a/pydantic/main.py#L49
IncEx: TypeAlias = "set[int] | set[str] | dict[int, Any] | dict[str, Any] | None"

PostParser = Callable[[Any], Any]



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Dict, List, Union, Optional
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = [
    "SubscriptionCreateParams",
    "PriceOverride",
    "PriceOverrideOverrideUnitPrice",
    "PriceOverrideOverrideUnitPriceUnitConfig",
    "PriceOverrideOverrideUnitPriceDiscount",
    "PriceOverrideOverridePackagePrice",
    "PriceOverrideOverridePackagePricePackageConfig",
    "PriceOverrideOverridePackagePriceDiscount",
    "PriceOverrideOverrideMatrixPrice",
    "PriceOverrideOverrideMatrixPriceMatrixConfig",
    "PriceOverrideOverrideMatrixPriceMatrixConfigMatrixValue",
    "PriceOverrideOverrideMatrixPriceDiscount",
    "PriceOverrideOverrideTieredPrice",
    "PriceOverrideOverrideTieredPriceTieredConfig",
    "PriceOverrideOverrideTieredPriceTieredConfigTier",
    "PriceOverrideOverrideTieredPriceDiscount",
    "PriceOverrideOverrideTieredBpsPrice",
    "PriceOverrideOverrideTieredBpsPriceTieredBpsConfig",
    "PriceOverrideOverrideTieredBpsPriceTieredBpsConfigTier",
    "PriceOverrideOverrideTieredBpsPriceDiscount",
    "PriceOverrideOverrideBpsPrice",
    "PriceOverrideOverrideBpsPriceBpsConfig",
    "PriceOverrideOverrideBpsPriceDiscount",
    "PriceOverrideOverrideBulkBpsPrice",
    "PriceOverrideOverrideBulkBpsPriceBulkBpsConfig",
    "PriceOverrideOverrideBulkBpsPriceBulkBpsConfigTier",
    "PriceOverrideOverrideBulkBpsPriceDiscount",
    "PriceOverrideOverrideBulkPrice",
    "PriceOverrideOverrideBulkPriceBulkConfig",
    "PriceOverrideOverrideBulkPriceBulkConfigTier",
    "PriceOverrideOverrideBulkPriceDiscount",
    "PriceOverrideOverrideTestRatingFunctionPrice",
    "PriceOverrideOverrideTestRatingFunctionPriceDiscount",
    "PriceOverrideOverrideFivetranExamplePrice",
    "PriceOverrideOverrideFivetranExamplePriceDiscount",
    "PriceOverrideOverrideThresholdTotalAmountPrice",
    "PriceOverrideOverrideThresholdTotalAmountPriceDiscount",
    "PriceOverrideOverrideTieredPackagePrice",
    "PriceOverrideOverrideTieredPackagePriceDiscount",
    "PriceOverrideOverrideTieredWithMinimumPrice",
    "PriceOverrideOverrideTieredWithMinimumPriceDiscount",
    "PriceOverrideOverridePackageWithAllocationPrice",
    "PriceOverrideOverridePackageWithAllocationPriceDiscount",
]


class SubscriptionCreateParams(TypedDict, total=False):
    align_billing_with_subscription_start_date: bool

    auto_collection: Optional[bool]

    aws_region: Optional[str]

    coupon_redemption_code: Optional[str]

    credits_overage_rate: Optional[float]

    customer_id: Optional[str]

    default_invoice_memo: Optional[str]

    end_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    external_customer_id: Optional[str]

    external_marketplace: Optional[Literal["google", "aws", "azure"]]

    external_marketplace_reporting_id: Optional[str]

    external_plan_id: Optional[str]
    """
    The external_plan_id of the plan that the given subscription should be switched
    to. Note that either this property or `plan_id` must be specified.
    """

    initial_phase_order: Optional[int]

    invoicing_threshold: Optional[str]

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    net_terms: Optional[int]

    per_credit_overage_amount: Optional[str]

    plan_id: Optional[str]
    """The plan that the given subscription should be switched to.

    Note that either this property or `external_plan_id` must be specified.
    """

    price_overrides: Optional[List[PriceOverride]]
    """Optionally provide a list of overrides for prices on the plan"""

    start_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]


class PriceOverrideOverrideUnitPriceUnitConfig(TypedDict, total=False):
    unit_amount: Required[str]
    """Rate per unit of usage"""

    scaling_factor: Optional[float]
    """Multiplier to scale rated quantity by"""


class PriceOverrideOverrideUnitPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideUnitPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["unit"]]

    unit_config: Required[PriceOverrideOverrideUnitPriceUnitConfig]

    discount: Optional[PriceOverrideOverrideUnitPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverridePackagePricePackageConfig(TypedDict, total=False):
    package_amount: Required[str]
    """A currency amount to rate usage by"""

    package_size: Optional[int]
    """An integer amount to represent package size.

    For example, 1000 here would divide usage by 1000 before multiplying by
    package_amount in rating
    """


class PriceOverrideOverridePackagePriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverridePackagePrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["package"]]

    package_config: Required[PriceOverrideOverridePackagePricePackageConfig]

    discount: Optional[PriceOverrideOverridePackagePriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideMatrixPriceMatrixConfigMatrixValue(TypedDict, total=False):
    dimension_values: Required[List[Optional[str]]]
    """One or two matrix keys to filter usage to this Matrix value by.

    For example, ["region", "tier"] could be used to filter cloud usage by a cloud
    region and an instance tier.
    """

    unit_amount: Required[str]
    """Unit price for the specified dimension_values"""

    scaling_factor: Optional[float]
    """Optional multiplier to scale rated quantities by"""


class PriceOverrideOverrideMatrixPriceMatrixConfig(TypedDict, total=False):
    default_unit_amount: Required[str]
    """Default per unit rate for any usage not bucketed into a specified matrix_value"""

    dimensions: Required[List[Optional[str]]]
    """One or two event property values to evaluate matrix groups by"""

    matrix_values: Required[List[PriceOverrideOverrideMatrixPriceMatrixConfigMatrixValue]]
    """Matrix values for specified matrix grouping keys"""

    scaling_factor: Optional[float]
    """
    Default optional multiplier to scale rated quantities that fall into the default
    bucket by
    """


class PriceOverrideOverrideMatrixPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideMatrixPrice(TypedDict, total=False):
    id: Required[str]

    matrix_config: Required[PriceOverrideOverrideMatrixPriceMatrixConfig]

    model_type: Required[Literal["matrix"]]

    discount: Optional[PriceOverrideOverrideMatrixPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredPriceTieredConfigTier(TypedDict, total=False):
    first_unit: Required[float]
    """Inclusive tier starting value"""

    unit_amount: Required[str]
    """Amount per unit"""

    last_unit: Optional[float]
    """Exclusive tier ending value. If null, this is treated as the last tier"""


class PriceOverrideOverrideTieredPriceTieredConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideTieredPriceTieredConfigTier]]
    """Tiers for rating based on total usage quantities into the specified tier"""


class PriceOverrideOverrideTieredPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered"]]

    tiered_config: Required[PriceOverrideOverrideTieredPriceTieredConfig]

    discount: Optional[PriceOverrideOverrideTieredPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredBpsPriceTieredBpsConfigTier(TypedDict, total=False):
    bps: Required[float]
    """Per-event basis point rate"""

    minimum_amount: Required[str]
    """Inclusive tier starting value"""

    maximum_amount: Optional[str]
    """Exclusive tier ending value"""

    per_unit_maximum: Optional[str]
    """Per unit maximum to charge"""


class PriceOverrideOverrideTieredBpsPriceTieredBpsConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideTieredBpsPriceTieredBpsConfigTier]]
    """
    Tiers for a Graduated BPS pricing model, where usage is bucketed into specified
    tiers
    """


class PriceOverrideOverrideTieredBpsPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredBpsPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered_bps"]]

    tiered_bps_config: Required[PriceOverrideOverrideTieredBpsPriceTieredBpsConfig]

    discount: Optional[PriceOverrideOverrideTieredBpsPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideBpsPriceBpsConfig(TypedDict, total=False):
    bps: Required[float]
    """Basis point take rate per event"""

    per_unit_maximum: Optional[str]
    """Optional currency amount maximum to cap spend per event"""


class PriceOverrideOverrideBpsPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideBpsPrice(TypedDict, total=False):
    id: Required[str]

    bps_config: Required[PriceOverrideOverrideBpsPriceBpsConfig]

    model_type: Required[Literal["bps"]]

    discount: Optional[PriceOverrideOverrideBpsPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideBulkBpsPriceBulkBpsConfigTier(TypedDict, total=False):
    bps: Required[float]
    """Basis points to rate on"""

    maximum_amount: Optional[str]
    """Upper bound for tier"""

    per_unit_maximum: Optional[str]
    """The maximum amount to charge for any one event"""


class PriceOverrideOverrideBulkBpsPriceBulkBpsConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideBulkBpsPriceBulkBpsConfigTier]]
    """
    Tiers for a bulk BPS pricing model where all usage is aggregated to a single
    tier based on total volume
    """


class PriceOverrideOverrideBulkBpsPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideBulkBpsPrice(TypedDict, total=False):
    id: Required[str]

    bulk_bps_config: Required[PriceOverrideOverrideBulkBpsPriceBulkBpsConfig]

    model_type: Required[Literal["bulk_bps"]]

    discount: Optional[PriceOverrideOverrideBulkBpsPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideBulkPriceBulkConfigTier(TypedDict, total=False):
    unit_amount: Required[str]
    """Amount per unit"""

    maximum_units: Optional[float]
    """Upper bound for this tier"""


class PriceOverrideOverrideBulkPriceBulkConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideBulkPriceBulkConfigTier]]
    """Bulk tiers for rating based on total usage volume"""


class PriceOverrideOverrideBulkPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideBulkPrice(TypedDict, total=False):
    id: Required[str]

    bulk_config: Required[PriceOverrideOverrideBulkPriceBulkConfig]

    model_type: Required[Literal["bulk"]]

    discount: Optional[PriceOverrideOverrideBulkPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTestRatingFunctionPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTestRatingFunctionPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["test_rating_function"]]

    test_rating_function_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideTestRatingFunctionPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideFivetranExamplePriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideFivetranExamplePrice(TypedDict, total=False):
    id: Required[str]

    fivetran_example_config: Required[Dict[str, object]]

    model_type: Required[Literal["fivetran_example"]]

    discount: Optional[PriceOverrideOverrideFivetranExamplePriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideThresholdTotalAmountPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideThresholdTotalAmountPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["threshold_total_amount"]]

    threshold_total_amount_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideThresholdTotalAmountPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredPackagePriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredPackagePrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered_package"]]

    tiered_package_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideTieredPackagePriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredWithMinimumPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredWithMinimumPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered_with_minimum"]]

    tiered_with_minimum_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideTieredWithMinimumPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverridePackageWithAllocationPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverridePackageWithAllocationPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["package_with_allocation"]]

    package_with_allocation_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverridePackageWithAllocationPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


PriceOverride = Union[
    PriceOverrideOverrideUnitPrice,
    PriceOverrideOverridePackagePrice,
    PriceOverrideOverrideMatrixPrice,
    PriceOverrideOverrideTieredPrice,
    PriceOverrideOverrideTieredBpsPrice,
    PriceOverrideOverrideBpsPrice,
    PriceOverrideOverrideBulkBpsPrice,
    PriceOverrideOverrideBulkPrice,
    PriceOverrideOverrideTestRatingFunctionPrice,
    PriceOverrideOverrideFivetranExamplePrice,
    PriceOverrideOverrideThresholdTotalAmountPrice,
    PriceOverrideOverrideTieredPackagePrice,
    PriceOverrideOverrideTieredWithMinimumPrice,
    PriceOverrideOverridePackageWithAllocationPrice,
]



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Any, List, Mapping, Optional, cast
from typing_extensions import Literal

import httpx

from ._utils import is_mapping

__all__ = [
    "BadRequestError",
    "AuthenticationError",
    "PermissionDeniedError",
    "NotFoundError",
    "ConflictError",
    "UnprocessableEntityError",
    "RateLimitError",
    "InternalServerError",
    "ConstraintViolation",
    "DuplicateResourceCreation",
    "FeatureNotAvailable",
    "RequestValidationError",
    "OrbAuthenticationError",
    "ResourceNotFound",
    "URLNotFound",
    "ResourceConflict",
    "RequestTooLarge",
    "ResourceTooLarge",
    "TooManyRequests",
    "OrbInternalServerError",
]


class OrbError(Exception):
    pass


class APIError(OrbError):
    message: str
    request: httpx.Request

    body: object | None
    """The API response body.

    If the API responded with a valid JSON structure then this property will be the
    decoded result.

    If it isn't a valid JSON structure then this will be the raw response.

    If there was no response associated with this error then it will be `None`.
    """

    def __init__(self, message: str, request: httpx.Request, *, body: object | None) -> None:  # noqa: ARG002
        super().__init__(message)
        self.request = request
        self.message = message


class APIResponseValidationError(APIError):
    response: httpx.Response
    status_code: int

    def __init__(self, response: httpx.Response, body: object | None, *, message: str | None = None) -> None:
        super().__init__(message or "Data returned by API invalid for expected schema.", response.request, body=body)
        self.response = response
        self.status_code = response.status_code


class APIStatusError(APIError):
    """Raised when an API response has a status code of 4xx or 5xx."""

    response: httpx.Response
    status_code: int

    def __init__(self, message: str, *, response: httpx.Response, body: object | None) -> None:
        super().__init__(message, response.request, body=body)
        self.response = response
        self.status_code = response.status_code


class APIConnectionError(APIError):
    def __init__(self, *, message: str = "Connection error.", request: httpx.Request) -> None:
        super().__init__(message, request, body=None)


class APITimeoutError(APIConnectionError):
    def __init__(self, request: httpx.Request) -> None:
        super().__init__(message="Request timed out.", request=request)


class BadRequestError(APIStatusError):
    status_code: Literal[400] = 400  # pyright: ignore[reportIncompatibleVariableOverride]


class AuthenticationError(APIStatusError):
    status_code: Literal[401] = 401  # pyright: ignore[reportIncompatibleVariableOverride]


class PermissionDeniedError(APIStatusError):
    status_code: Literal[403] = 403  # pyright: ignore[reportIncompatibleVariableOverride]


class NotFoundError(APIStatusError):
    status_code: Literal[404] = 404  # pyright: ignore[reportIncompatibleVariableOverride]


class ConflictError(APIStatusError):
    status_code: Literal[409] = 409  # pyright: ignore[reportIncompatibleVariableOverride]


class UnprocessableEntityError(APIStatusError):
    status_code: Literal[422] = 422  # pyright: ignore[reportIncompatibleVariableOverride]


class RateLimitError(APIStatusError):
    status_code: Literal[429] = 429  # pyright: ignore[reportIncompatibleVariableOverride]


class InternalServerError(APIStatusError):
    pass


class ConstraintViolation(BadRequestError):
    status: Literal[400]

    type: Literal["https://docs.withorb.com/reference/error-responses#400-constraint-violation"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class DuplicateResourceCreation(BadRequestError):
    status: Literal[400]

    type: Literal["https://docs.withorb.com/reference/error-responses#400-duplicate-resource-creation"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class FeatureNotAvailable(BadRequestError):
    status: Literal[400]

    type: Literal["https://docs.withorb.com/reference/error-responses#404-feature-not-available"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class RequestValidationError(BadRequestError):
    status: Literal[400]

    type: Literal["https://docs.withorb.com/reference/error-responses#400-request-validation-errors"]

    validation_errors: List[object]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.validation_errors = cast(Any, data.get("validation_errors"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class OrbAuthenticationError(AuthenticationError):
    status: Literal[401]

    type: Literal["https://docs.withorb.com/reference/error-responses#401-authentication-error"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class ResourceNotFound(NotFoundError):
    status: Literal[404]

    title: str

    type: Literal["https://docs.withorb.com/reference/error-responses#404-resource-not-found"]

    detail: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.title = cast(Any, data.get("title"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))


class URLNotFound(NotFoundError):
    status: Literal[404]

    type: Literal["https://docs.withorb.com/reference/error-responses#404-url-not-found"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class ResourceConflict(ConflictError):
    status: Literal[409]

    type: Literal["https://docs.withorb.com/reference/error-responses#409-resource-conflict"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class RequestTooLarge(APIStatusError):
    status: Literal[413]

    type: Literal["https://docs.withorb.com/reference/error-responses#413-request-too-large"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class ResourceTooLarge(APIStatusError):
    status: Literal[413]

    type: Literal["https://docs.withorb.com/reference/error-responses#413-resource-too-large"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class TooManyRequests(RateLimitError):
    status: Literal[429]

    type: Literal["https://docs.withorb.com/reference/error-responses#429-too-many-requests"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))


class OrbInternalServerError(InternalServerError):
    status: Literal[500]

    type: Literal["https://docs.withorb.com/reference/error-responses#500-internal-server-error"]

    detail: Optional[str] = None

    title: Optional[str] = None

    def __init__(self, message: str, *, body: object, response: httpx.Response) -> None:
        data = cast(Mapping[str, object], body if is_mapping(body) else {})
        super().__init__(message, response=response, body=body)

        self.status = cast(Any, data.get("status"))
        self.type = cast(Any, data.get("type"))
        self.detail = cast(Any, data.get("detail"))
        self.title = cast(Any, data.get("title"))



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

import os
import asyncio
from typing import Any, Union, Mapping
from typing_extensions import Self, override

import httpx

from . import resources, _exceptions
from ._qs import Querystring
from ._types import (
    NOT_GIVEN,
    Omit,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
)
from ._utils import is_given, is_mapping, get_async_library
from ._version import __version__
from ._streaming import Stream as Stream
from ._streaming import AsyncStream as AsyncStream
from ._exceptions import OrbError, APIStatusError
from ._base_client import DEFAULT_MAX_RETRIES, SyncAPIClient, AsyncAPIClient

__all__ = [
    "Timeout",
    "Transport",
    "ProxiesTypes",
    "RequestOptions",
    "resources",
    "Orb",
    "AsyncOrb",
    "Client",
    "AsyncClient",
]


class Orb(SyncAPIClient):
    top_level: resources.TopLevel
    coupons: resources.Coupons
    credit_notes: resources.CreditNotes
    customers: resources.Customers
    events: resources.Events
    invoice_line_items: resources.InvoiceLineItems
    invoices: resources.Invoices
    items: resources.Items
    metrics: resources.Metrics
    plans: resources.Plans
    prices: resources.Prices
    subscriptions: resources.Subscriptions
    with_raw_response: OrbWithRawResponse

    # client options
    api_key: str

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client. See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous orb client instance.

        This automatically infers the `api_key` argument from the `ORB_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("ORB_API_KEY")
        if api_key is None:
            raise OrbError(
                "The api_key client option must be set either by passing api_key to the client or by setting the ORB_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("ORB_BASE_URL")
        if base_url is None:
            base_url = f"https://api.withorb.com/v1"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._idempotency_header = "Idempotency-Key"

        self.top_level = resources.TopLevel(self)
        self.coupons = resources.Coupons(self)
        self.credit_notes = resources.CreditNotes(self)
        self.customers = resources.Customers(self)
        self.events = resources.Events(self)
        self.invoice_line_items = resources.InvoiceLineItems(self)
        self.invoices = resources.Invoices(self)
        self.items = resources.Items(self)
        self.metrics = resources.Metrics(self)
        self.plans = resources.Plans(self)
        self.prices = resources.Prices(self)
        self.subscriptions = resources.Subscriptions(self)
        self.with_raw_response = OrbWithRawResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or str(self.base_url),
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def __del__(self) -> None:
        if not hasattr(self, "_has_custom_http_client") or not hasattr(self, "close"):
            # this can happen if the '__init__' method raised an error
            return

        if self._has_custom_http_client:
            return

        self.close()

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        type_ = body.get("type") if is_mapping(body) else None
        if type_ == "https://docs.withorb.com/reference/error-responses#400-constraint-violation":
            return _exceptions.ConstraintViolation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-duplicate-resource-creation":
            return _exceptions.DuplicateResourceCreation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-feature-not-available":
            return _exceptions.FeatureNotAvailable(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-request-validation-errors":
            return _exceptions.RequestValidationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#401-authentication-error":
            return _exceptions.OrbAuthenticationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-resource-not-found":
            return _exceptions.ResourceNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-url-not-found":
            return _exceptions.URLNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#409-resource-conflict":
            return _exceptions.ResourceConflict(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-request-too-large":
            return _exceptions.RequestTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-resource-too-large":
            return _exceptions.ResourceTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#429-too-many-requests":
            return _exceptions.TooManyRequests(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#500-internal-server-error":
            return _exceptions.OrbInternalServerError(err_msg, response=response, body=body)
        if response.status_code >= 500:
            return _exceptions.OrbInternalServerError(
                err_msg,
                response=response,
                body={
                    "status": 500,
                    "type": "https://docs.withorb.com/reference/error-responses#500-internal-server-error",
                    "detail": None,
                    "title": None,
                },
            )

        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncOrb(AsyncAPIClient):
    top_level: resources.AsyncTopLevel
    coupons: resources.AsyncCoupons
    credit_notes: resources.AsyncCreditNotes
    customers: resources.AsyncCustomers
    events: resources.AsyncEvents
    invoice_line_items: resources.AsyncInvoiceLineItems
    invoices: resources.AsyncInvoices
    items: resources.AsyncItems
    metrics: resources.AsyncMetrics
    plans: resources.AsyncPlans
    prices: resources.AsyncPrices
    subscriptions: resources.AsyncSubscriptions
    with_raw_response: AsyncOrbWithRawResponse

    # client options
    api_key: str

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client. See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async orb client instance.

        This automatically infers the `api_key` argument from the `ORB_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("ORB_API_KEY")
        if api_key is None:
            raise OrbError(
                "The api_key client option must be set either by passing api_key to the client or by setting the ORB_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("ORB_BASE_URL")
        if base_url is None:
            base_url = f"https://api.withorb.com/v1"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._idempotency_header = "Idempotency-Key"

        self.top_level = resources.AsyncTopLevel(self)
        self.coupons = resources.AsyncCoupons(self)
        self.credit_notes = resources.AsyncCreditNotes(self)
        self.customers = resources.AsyncCustomers(self)
        self.events = resources.AsyncEvents(self)
        self.invoice_line_items = resources.AsyncInvoiceLineItems(self)
        self.invoices = resources.AsyncInvoices(self)
        self.items = resources.AsyncItems(self)
        self.metrics = resources.AsyncMetrics(self)
        self.plans = resources.AsyncPlans(self)
        self.prices = resources.AsyncPrices(self)
        self.subscriptions = resources.AsyncSubscriptions(self)
        self.with_raw_response = AsyncOrbWithRawResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or str(self.base_url),
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def __del__(self) -> None:
        if not hasattr(self, "_has_custom_http_client") or not hasattr(self, "close"):
            # this can happen if the '__init__' method raised an error
            return

        if self._has_custom_http_client:
            return

        try:
            asyncio.get_running_loop().create_task(self.close())
        except Exception:
            pass

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        type_ = body.get("type") if is_mapping(body) else None
        if type_ == "https://docs.withorb.com/reference/error-responses#400-constraint-violation":
            return _exceptions.ConstraintViolation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-duplicate-resource-creation":
            return _exceptions.DuplicateResourceCreation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-feature-not-available":
            return _exceptions.FeatureNotAvailable(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-request-validation-errors":
            return _exceptions.RequestValidationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#401-authentication-error":
            return _exceptions.OrbAuthenticationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-resource-not-found":
            return _exceptions.ResourceNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-url-not-found":
            return _exceptions.URLNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#409-resource-conflict":
            return _exceptions.ResourceConflict(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-request-too-large":
            return _exceptions.RequestTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-resource-too-large":
            return _exceptions.ResourceTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#429-too-many-requests":
            return _exceptions.TooManyRequests(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#500-internal-server-error":
            return _exceptions.OrbInternalServerError(err_msg, response=response, body=body)
        if response.status_code >= 500:
            return _exceptions.OrbInternalServerError(
                err_msg,
                response=response,
                body={
                    "status": 500,
                    "type": "https://docs.withorb.com/reference/error-responses#500-internal-server-error",
                    "detail": None,
                    "title": None,
                },
            )

        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class OrbWithRawResponse:
    def __init__(self, client: Orb) -> None:
        self.top_level = resources.TopLevelWithRawResponse(client.top_level)
        self.coupons = resources.CouponsWithRawResponse(client.coupons)
        self.credit_notes = resources.CreditNotesWithRawResponse(client.credit_notes)
        self.customers = resources.CustomersWithRawResponse(client.customers)
        self.events = resources.EventsWithRawResponse(client.events)
        self.invoice_line_items = resources.InvoiceLineItemsWithRawResponse(client.invoice_line_items)
        self.invoices = resources.InvoicesWithRawResponse(client.invoices)
        self.items = resources.ItemsWithRawResponse(client.items)
        self.metrics = resources.MetricsWithRawResponse(client.metrics)
        self.plans = resources.PlansWithRawResponse(client.plans)
        self.prices = resources.PricesWithRawResponse(client.prices)
        self.subscriptions = resources.SubscriptionsWithRawResponse(client.subscriptions)


class AsyncOrbWithRawResponse:
    def __init__(self, client: AsyncOrb) -> None:
        self.top_level = resources.AsyncTopLevelWithRawResponse(client.top_level)
        self.coupons = resources.AsyncCouponsWithRawResponse(client.coupons)
        self.credit_notes = resources.AsyncCreditNotesWithRawResponse(client.credit_notes)
        self.customers = resources.AsyncCustomersWithRawResponse(client.customers)
        self.events = resources.AsyncEventsWithRawResponse(client.events)
        self.invoice_line_items = resources.AsyncInvoiceLineItemsWithRawResponse(client.invoice_line_items)
        self.invoices = resources.AsyncInvoicesWithRawResponse(client.invoices)
        self.items = resources.AsyncItemsWithRawResponse(client.items)
        self.metrics = resources.AsyncMetricsWithRawResponse(client.metrics)
        self.plans = resources.AsyncPlansWithRawResponse(client.plans)
        self.prices = resources.AsyncPricesWithRawResponse(client.prices)
        self.subscriptions = resources.AsyncSubscriptionsWithRawResponse(client.subscriptions)


Client = Orb

AsyncClient = AsyncOrb



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from ..._utils import PropertyInfo

__all__ = ["BackfillCreateParams"]


class BackfillCreateParams(TypedDict, total=False):
    timeframe_end: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """The (exclusive) end of the usage timeframe affected by this backfill."""

    timeframe_start: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """The (inclusive) start of the usage timeframe affected by this backfill."""

    close_time: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """The time at which no more events will be accepted for this backfill.

    The backfill will automatically begin reflecting throughout Orb at the close
    time. If not specified, it will default to 1 day after the creation of the
    backfill.
    """

    customer_id: Optional[str]
    """The ID of the customer to which this backfill is scoped."""

    external_customer_id: Optional[str]
    """The external customer ID of the customer to which this backfill is scoped."""

    replace_existing_events: bool
    """
    If true, replaces all existing events in the timeframe with the newly ingested
    events. If false, adds the newly ingested events to the existing events.
    """



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Union
from datetime import date
from typing_extensions import Literal, Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["SubscriptionUpdateFixedFeeQuantityParams"]


class SubscriptionUpdateFixedFeeQuantityParams(TypedDict, total=False):
    price_id: Required[str]
    """Price for which the quantity should be updated. Must be a fixed fee."""

    quantity: Required[float]

    change_option: Literal["immediate", "upcoming_invoice", "effective_date"]
    """Determines when the change takes effect.

    Note that if `effective_date` is specified, this defaults to `effective_date`.
    Otherwise, this defaults to `immediate` unless it's explicitly set to
    `upcoming_invoice.
    """

    effective_date: Annotated[Union[str, date, None], PropertyInfo(format="iso8601")]
    """
    The date that the quantity change should take effect, localized to the
    customer's timezone. Ifthis parameter is not passed in, the quantity change is
    effective according to `change_option`.
    """



# File generated from our OpenAPI spec by Stainless.

from typing import List, Generic, TypeVar, Optional
from typing_extensions import override

from ._types import ModelT
from ._models import BaseModel
from ._base_client import BasePage, PageInfo, BaseSyncPage, BaseAsyncPage

__all__ = ["SyncPage", "AsyncPage"]

_BaseModelT = TypeVar("_BaseModelT", bound=BaseModel)


class PagePaginationMetadata(BaseModel):
    has_more: bool

    next_cursor: Optional[str]


class SyncPage(BaseSyncPage[ModelT], BasePage[ModelT], Generic[ModelT]):
    data: List[ModelT]
    pagination_metadata: PagePaginationMetadata

    @override
    def _get_page_items(self) -> List[ModelT]:
        return self.data

    @override
    def next_page_info(self) -> Optional[PageInfo]:
        cursor = self.pagination_metadata.next_cursor
        if not cursor:
            return None

        return PageInfo(params={"cursor": cursor})


class AsyncPage(BaseAsyncPage[ModelT], BasePage[ModelT], Generic[ModelT]):
    data: List[ModelT]
    pagination_metadata: PagePaginationMetadata

    @override
    def _get_page_items(self) -> List[ModelT]:
        return self.data

    @override
    def next_page_info(self) -> Optional[PageInfo]:
        cursor = self.pagination_metadata.next_cursor
        if not cursor:
            return None

        return PageInfo(params={"cursor": cursor})
